/*
 * gsampler-channel-manager.c
 * Copyright (C) Lilian Monchalin 2011 <isaya07@users.sf.net>
 * 
 * gsampler is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * gsampler is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <glib/gi18n-lib.h>

#include "gsampler-channel-manager.h"
#include "gsampler-window.h"
#include "gsampler-utils.h"
#include "gsampler-channel.h"
#include "gsampler-client-channel.h"
#include "gsampler-client-device.h"

G_DEFINE_TYPE (GsamplerChannelManager, gsampler_channel_manager, GTK_TYPE_BOX);

enum {
	CHANNELS_UPDATE_EVENT,
	CHANNELS_NB_SIGNALS
};

static guint channels_signals[CHANNELS_NB_SIGNALS] = { 0 };

static GsamplerChannelManager* gsampler_channel_manager__instance = NULL;
static gboolean gsampler_channel_manager_status = FALSE;
static GSList *channel_list = NULL;

static gboolean 
gsampler_channel_manager_check_device (GsamplerChannelManager* self) 
{
	gboolean result = TRUE;
	GsamplerClient* client;
	gint audio;
	gint midi;
	g_return_val_if_fail (self != NULL, FALSE);
	
	client = gsampler_client_get_instance ();
	audio = gsampler_client_get_audio_output_count (client);
	midi = gsampler_client_get_midi_input_count (client);
	if (audio == 0 || midi == 0) 
	{
		result = FALSE;
	}
	g_object_unref (client);
	return result;
}

static void 
gsampler_channel_manager_on_channel_add (GtkButton* sender, gpointer data) 
{
	GsamplerChannelManager* self = (GsamplerChannelManager*) data;
	g_return_if_fail (GSAMPLER_IS_CHANNEL_MANAGER (self));
	g_return_if_fail (sender != NULL);
	
	if (gsampler_channel_manager_status == TRUE) {
		gsampler_channel_manager_create_channel (self);
	}
}

static void
gsampler_channel_manager_on_channel_updated (GsamplerChannel* sender, gpointer data) 
{
	GsamplerChannelManager* self = (GsamplerChannelManager*) data;
	g_return_if_fail (GSAMPLER_IS_CHANNEL_MANAGER (self));

	g_signal_emit (self, channels_signals[CHANNELS_UPDATE_EVENT], 0);
}

static gboolean 
gsampler_channel_manager_create_channel_form (GsamplerChannelManager* self, gint chan_id) 
{
	GsamplerChannel* channel = NULL;
	g_return_val_if_fail (self != NULL, FALSE);

	if (g_id_list_exist (channel_list, chan_id)) 
	{
		return TRUE;
	}
	channel = gsampler_channel_new ();
	if (!gsampler_channel_active (channel, chan_id)) 
	{
		gsampler_main_window_send_message ("warning", g_strdup_printf ("Cant get info for channel : %d", chan_id));
		gtk_widget_destroy ((GtkWidget*) channel);
		return FALSE;
	} else
	{
		channel_list = g_id_list_add (channel_list, chan_id, g_object_ref (channel));
		g_signal_connect_object (channel, "channel-updated", 
	                      		 (GCallback) gsampler_channel_manager_on_channel_updated, self, 0);
		gtk_box_pack_start (self->priv->channel_box, (GtkWidget*) channel, FALSE, FALSE, 0);
		gtk_widget_show ((GtkWidget*) channel);
	}
	return TRUE;
}

static void 
gsampler_channel_manager_delete_channel_form (GsamplerChannelManager* self, gint id)
{
	GsamplerChannel* channel = NULL;
	g_return_if_fail (self != NULL);

	channel = (GsamplerChannel*) g_id_list_get (channel_list, id);
	if (GSAMPLER_IS_CHANNEL (channel))
	{
		channel_list = g_id_list_remove (channel_list, id);
		gtk_widget_destroy ((GtkWidget*) channel);
	} else
	{
		gsampler_main_window_send_message ("warning",
		                                   g_strdup_printf ("Can't delete channel form for channel : %d", id));
	}
}

static void 
gsampler_channel_manager_on_voice_event (GsamplerClient* sender, gint id, 
                                         gint val, gpointer data) 
{
	GSList *list = NULL;
	GsamplerChannelManager* self = (GsamplerChannelManager*) data;
	g_return_if_fail (GSAMPLER_IS_CHANNEL_MANAGER (self));

	list = channel_list;
	if (g_id_list_exist (list, id)) 
	{
		GsamplerChannel* channel;
		channel = (GsamplerChannel*) g_id_list_get (list, id);
		gsampler_channel_update_voice_count (channel, val);
	}
}

static void 
gsampler_channel_manager_on_stream_event (GsamplerClient* sender, gint id, 
                                         gint val, gpointer data) 
{
	GSList* list = NULL;
	GsamplerChannelManager* self = (GsamplerChannelManager*) data;
	g_return_if_fail (GSAMPLER_IS_CHANNEL_MANAGER (self));

	list = channel_list;
	if (g_id_list_exist (list, id)) 
	{
		GsamplerChannel* channel;
		channel = (GsamplerChannel*) g_id_list_get (list, id);
		gsampler_channel_update_stream_count (channel, val);
	}
}

static void 
gsampler_channel_manager_on_info_event (GsamplerClient* sender, gint id, gpointer data) 
{
	GSList* list = NULL;
	GsamplerChannelManager* self = (GsamplerChannelManager*) data;
	g_return_if_fail (GSAMPLER_IS_CHANNEL_MANAGER (self));

	list = channel_list;
	if (g_id_list_exist (list, id)) 
	{
		GsamplerChannel* channel;
		channel = (GsamplerChannel*) g_id_list_get (list, id);
		gsampler_channel_update_info (channel);
	}
}

static void 
gsampler_channel_manager_on_update_event (GsamplerClient* sender, gchar** chanlist, gpointer data) 
{
	GSList *list = NULL;
	gint chanlist_length = 0;
	GsamplerChannelManager* self = (GsamplerChannelManager*) data;
	g_return_if_fail (GSAMPLER_IS_CHANNEL_MANAGER (self));
											  
	while (!gsampler_channel_manager_status) 
	{
		g_usleep ((guint) 300000);
	}
	gsampler_channel_manager_status = FALSE;

	if (chanlist != NULL)
		chanlist_length = g_strv_length (chanlist);
	list = channel_list;
	while (list) 
	{
		gint chan_id, dev_it;
		gboolean result = TRUE;
		chan_id = ((struct Struct_list_widget*) (list->data))->Sid;
		for (dev_it = 0; dev_it < chanlist_length; dev_it = dev_it + 1) {
			gchar* dev;
			dev = g_strdup (chanlist[dev_it]);
			if (g_strcmp0 (dev, "NONE") == 0) 
			{
				result = TRUE;
				g_free (dev);
				break;
			} else 
			{
				gint new_id;
				new_id = atoi (dev);
				if (chan_id == new_id) 
				{
					result = FALSE;
				}
			}
			g_free (dev);
		}
		if (result) 
		{
			gsampler_channel_manager_delete_channel_form (self, chan_id);
		}
		list = g_slist_next(list);
	}
	int dev_it;
	for (dev_it = 0; dev_it < chanlist_length; dev_it = dev_it + 1) 
	{
		gchar* dev;
		dev = g_strdup (chanlist[dev_it]);
		if (g_strcmp0 (dev, "NONE") != 0) 
		{
			gint chan_id;
			chan_id = atoi (dev);
			if (!g_id_list_exist (list, chan_id)) 
			{
				if (!gsampler_channel_manager_create_channel_form (self, chan_id)) 
				{
					gsampler_main_window_send_message ("warning", 
							                           g_strdup_printf ("Can't create channel form : %i", chan_id));	
				}
			}	
		}
		g_free (dev);
	}
	gsampler_channel_manager_status = TRUE;
	g_signal_emit (self, channels_signals [CHANNELS_UPDATE_EVENT], 0);
}

static GObject * 
gsampler_channel_manager_constructor (GType type, guint n_construct_properties, 
                               		  GObjectConstructParam * construct_properties) 
{
	GObject * obj;
	GsamplerChannelManager * self = NULL;
	GtkBox* hbox1 = NULL;
	GtkLabel* lbl = NULL;
	GtkButton* button = NULL;
	GtkScrolledWindow* scrol = NULL;
								   
	obj = G_OBJECT_CLASS (gsampler_channel_manager_parent_class)->constructor (type, n_construct_properties, construct_properties);
	self = GSAMPLER_CHANNEL_MANAGER (obj);
	//gtk_box_set_spacing ((GtkBox*) self, 5);
	gtk_container_set_border_width ((GtkContainer*) self, 2);
	gtk_orientable_set_orientation ((GtkOrientable*) self, GTK_ORIENTATION_VERTICAL);
	hbox1 = (GtkBox*) gtk_hbox_new (FALSE, 5);
#if GTK_CHECK_VERSION (2, 90, 0)
	gtk_widget_set_margin_bottom ((GtkWidget*) hbox1, 3);
	gtk_widget_set_margin_top ((GtkWidget*) hbox1, 3);
#endif
	lbl = (GtkLabel*) gtk_label_new (NULL);
	gtk_label_set_markup ((GtkLabel *) lbl, _("<b>Channels :</b>"));
	button = (GtkButton*) gtk_button_new_from_stock (GTK_STOCK_ADD);
	scrol = (GtkScrolledWindow*) gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (scrol, GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type (scrol, GTK_SHADOW_IN);
	GtkWidget *viewport = gtk_viewport_new (gtk_scrolled_window_get_hadjustment (scrol),
	                                         gtk_scrolled_window_get_vadjustment (scrol));
	gtk_viewport_set_shadow_type ((GtkViewport *) viewport, GTK_SHADOW_NONE);
		
	self->priv->channel_box = (GtkBox*) gtk_vbox_new (FALSE, 0);
	gtk_box_pack_start (hbox1, GTK_WIDGET (lbl), TRUE, TRUE, (guint) 5);
	gtk_box_pack_start (hbox1, GTK_WIDGET (button), FALSE, TRUE, (guint) 5);
	gtk_container_add((GtkContainer*) scrol, viewport);
	gtk_container_add ((GtkContainer*) viewport, (GtkWidget*) self->priv->channel_box);
	gtk_box_pack_start ((GtkBox*) self, GTK_WIDGET (hbox1), FALSE, TRUE, (guint) 3);
	gtk_box_pack_start ((GtkBox*) self, GTK_WIDGET (scrol), TRUE, TRUE, (guint) 0);
								   
	g_signal_connect_object (button, "clicked", 
	                         (GCallback) gsampler_channel_manager_on_channel_add, self, 0);
								   
	gtk_widget_show_all ((GtkWidget*) self);
	gtk_widget_set_sensitive ((GtkWidget*) self, FALSE);
	return obj;
}

static void gsampler_channel_manager_init (GsamplerChannelManager * self) 
{
	self->priv = GSAMPLER_CHANNEL_MANAGER_GET_PRIVATE (self);
}

static void 
gsampler_channel_manager_finalize (GObject* obj) 
{
	GsamplerChannelManager * self;
	self = GSAMPLER_CHANNEL_MANAGER (obj);
	g_object_unref (self->priv->channel_box);
	G_OBJECT_CLASS (gsampler_channel_manager_parent_class)->finalize (obj);
}

static void 
gsampler_channel_manager_class_init (GsamplerChannelManagerClass * klass) 
{
	gsampler_channel_manager_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (GsamplerChannelManagerPrivate));
	G_OBJECT_CLASS (klass)->constructor = gsampler_channel_manager_constructor;
	G_OBJECT_CLASS (klass)->finalize = gsampler_channel_manager_finalize;

	channels_signals[CHANNELS_UPDATE_EVENT] = g_signal_new ("channels_updated", 
	                                                        GSAMPLER_TYPE_CHANNEL_MANAGER, 
	                                                        G_SIGNAL_RUN_LAST, 
	                                                        0, NULL, NULL, 
	                                                        g_cclosure_marshal_VOID__VOID, 
	                                                        G_TYPE_NONE, 0);
}

GsamplerChannelManager* 
gsampler_channel_manager_new (void) 
{
	return g_object_new (GSAMPLER_TYPE_CHANNEL_MANAGER, NULL);
}

GsamplerChannelManager* 
gsampler_channel_manager_get_instance (void) {
	if (gsampler_channel_manager__instance == NULL) 
	{
		gsampler_channel_manager__instance = gsampler_channel_manager_new ();
	}
	return g_object_ref (gsampler_channel_manager__instance);
}

GSList* 
gsampler_channel_manager_get_channels (void) 
{
	return channel_list;
}

gint 
gsampler_channel_manager_create_channel (GsamplerChannelManager* self) 
{
	GsamplerClient* client = NULL;
	gint id = -1;
	gchar** audio_list = NULL;
	gchar** midi_list = NULL;
	g_return_val_if_fail (self != NULL, -1);
	
	client = gsampler_client_get_instance ();
	if (!gsampler_channel_manager_check_device (self)) 
	{
		gsampler_main_window_send_message ("info", _("Create audio/midi device first"));
	} else 
	{
		audio_list = gsampler_client_get_available_audio_output (client);
		midi_list = gsampler_client_get_available_midi_input (client);
		id = gsampler_client_create_channel (client);
		if (id == (-1)) 
		{
			gsampler_main_window_send_message ("warning", _("Error can't create channel"));
		} else if (!gsampler_client_set_channel_output (client, id, atoi (audio_list[0]))) 
		{
			id = -1;
			gsampler_main_window_send_message ("warning", _("Error can't connect to audio device"));
		} else if (!gsampler_client_set_channel_input (client, id, atoi (midi_list[0]))) 
		{
			id = -1;
			gsampler_main_window_send_message ("warning", _("Error can't connect to midi device"));
		} else if (!gsampler_client_load_engine (client, "GIG", id)) 
		{
			id = -1;
			gsampler_main_window_send_message ("warning", _("Error can't load engine"));
		} 
	}
	g_strfreev (midi_list);
	g_strfreev (audio_list);
	g_object_unref (client);
	return id;
}


gboolean 
gsampler_channel_manager_update_channel (GsamplerChannelManager* self, gint id, 
                                         const gchar* engine, const gchar* file, gint* inst_id) 
{
	GsamplerClient* client;
	g_return_val_if_fail (self != NULL, FALSE);
	
	client = gsampler_client_get_instance ();
	gsampler_main_window_set_wait_cursor (TRUE);
	if (engine != NULL) {
		if (!gsampler_client_load_engine (client, engine, id)) 
		{
			gsampler_main_window_set_wait_cursor (FALSE);
			gsampler_main_window_send_message ("warning", 
			                                   g_strdup_printf ("Error can't load engine for channel : %d", id));
			g_object_unref (client);
			return FALSE;
		}
	}
	if (file != NULL && inst_id != NULL) 
	{
		if (!gsampler_client_load_instrument (client, id, file, *inst_id)) 
		{
			gsampler_main_window_set_wait_cursor (FALSE);
			gsampler_main_window_send_message ("warning", 
			                                   g_strdup_printf ("Error can't load instrument for channel : %d", id));
			
			g_object_unref (client);
			return FALSE;
		}
	}
	g_object_unref (client);
	return TRUE;
}

void gsampler_channel_manager_active (GsamplerChannelManager* self) 
{
	GsamplerClient* client = NULL;
	gchar** channels;
	g_return_if_fail (self != NULL);
	
	client = gsampler_client_get_instance ();
	gsampler_channel_manager_status = TRUE;
	channels = gsampler_client_get_available_channel (client);
	gsampler_channel_manager_on_update_event (client, channels, self);
	g_signal_connect_object (client, "channel-count-event",
	                         (GCallback) gsampler_channel_manager_on_update_event, self, 0);
	g_signal_connect_object (client, "channel-stream-event", 
	                         (GCallback) gsampler_channel_manager_on_stream_event, self, 0);
	g_signal_connect_object (client, "channel-voice-event", 
	                         (GCallback) gsampler_channel_manager_on_voice_event, self, 0);
	g_signal_connect_object (client, "channel-info-event", 
	                         (GCallback) gsampler_channel_manager_on_info_event, self, 0);
	gtk_widget_set_sensitive ((GtkWidget*) self, TRUE);
	g_strfreev (channels);
	g_object_unref (client);
}


void gsampler_channel_manager_inactive (GsamplerChannelManager* self) 
{
	GsamplerClient* client = NULL;
	g_return_if_fail (self != NULL);
	
	client = gsampler_client_get_instance ();
	gsampler_channel_manager_status = FALSE;
	g_signal_handlers_disconnect_by_func (client, 
	                                      gsampler_channel_manager_on_update_event, self);
	g_signal_handlers_disconnect_by_func (client, 
	                                      gsampler_channel_manager_on_stream_event, self);
	g_signal_handlers_disconnect_by_func (client, 
	                                      gsampler_channel_manager_on_voice_event, self);
	g_signal_handlers_disconnect_by_func (client, 
	                                      gsampler_channel_manager_on_info_event, self);
	gtk_widget_set_sensitive ((GtkWidget*) self, FALSE);
	g_object_unref (client);
}

void 
gsampler_channel_manager_on_note_on (GsamplerKeyboard* sender, gint chan,
                                     gint pitch, gint vel, gpointer data) 
{
	GSList *list = NULL;
	GsamplerChannelManager* self = GSAMPLER_CHANNEL_MANAGER (data);
	g_return_if_fail (GSAMPLER_CHANNEL_MANAGER (self));
	
	list = channel_list;	
	if (g_id_list_exist (list, chan)) 
	{
		GsamplerClient* client;
		client = gsampler_client_get_instance ();
		gsampler_client_send_channel_midi_note_on (client, chan, pitch, vel);
		g_object_unref (client);
	}
}


void 
gsampler_channel_manager_on_note_off (GsamplerKeyboard* sender, gint chan,
                                      gint pitch, gint vel, gpointer data) 
{
	GSList *list = NULL;
	GsamplerChannelManager* self = GSAMPLER_CHANNEL_MANAGER (data);
	g_return_if_fail (GSAMPLER_CHANNEL_MANAGER (self));
	
	list = channel_list;	
	if (g_id_list_exist (list, chan)) 
	{
		GsamplerClient* client;
		client = gsampler_client_get_instance ();
		gsampler_client_send_channel_midi_note_off (client, chan, pitch, vel);
		g_object_unref (client);
	}
}
